"""Color constants for use in image processing and graphical user interfaces."""

# Standard RGB colors
RED = (255, 0, 0)
GREEN = (0, 255, 0)
BLUE = (0, 0, 255)

# OpenCV colors w/ color order BGR
RED_CV = (0, 0, 255)
GREEN_CV = (0, 255, 0)
BLUE_CV = (255, 0, 0)

# Colors
BLACK = 0
BLACK3 = (BLACK,) * 3
WHITE = 255
WHITE3 = (WHITE,) * 3
